"""UvT specific extensions to the authentication backend"""

from datetime import datetime, timezone, timedelta
from functools import cached_property
import os.path

import ldap.filter

from django.conf import settings as django_settings
from django.contrib.auth.backends import RemoteUserBackend as DjangoRemoteUserBackend
from django.db.transaction import atomic, TransactionManagementError


class RemoteUserBackend(DjangoRemoteUserBackend):
    """UvT specific extensions to the authentication backend"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__settings = django_settings.UVT_USER_BACKEND

    @cached_property
    def ldap_connection(self):
        settings = self.__settings
        connection = ldap.ldapobject.ReconnectLDAPObject(settings['ldap_server_uri'])
        connection.set_option(ldap.OPT_NETWORK_TIMEOUT, 5)
        ca_cert_path = settings.get('ca_cert_path')
        if ca_cert_path is None:
            connection.set_option(ldap.OPT_X_TLS_CACERTDIR, '/etc/ssl/certs')
            connection.set_option(
                ldap.OPT_X_TLS_CACERTFILE, '/etc/ssl/certs/ca-certificates.crt'
            )
        else:
            if os.path.isdir(ca_cert_path):
                connection.set_option(ldap.OPT_X_TLS_CACERTDIR, ca_cert_path)
            else:
                connection.set_option(ldap.OPT_X_TLS_CACERTFILE, ca_cert_path)
        connection.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_DEMAND)
        # apply the SSL configuration
        connection.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
        connection.start_tls_s()
        return connection

    @cached_property
    def _ldap_to_model(self):
        configured = self.__settings['attributes']
        return {key.lower(): value for key, value in configured.items()}

    def _update_user_using_ldap(self, user):
        ldap_to_model = self._ldap_to_model
        settings = self.__settings

        connection = self.ldap_connection
        connection.simple_bind_s(settings.get('bind_dn'), settings.get('bind_password'))
        results = connection.search_s(
            settings['search_base'],
            settings.get('search_scope', ldap.SCOPE_SUBTREE),
            ldap.filter.filter_format(
                settings.get('search_filter', '(uid=%s)'), (user.username,)
            ),
            list(ldap_to_model.keys()),
        )

        # It is unspecified in the LDAP docs whether results is an iterator
        # or a list. This code will work in either case:
        attrs = None
        for result in results:
            if attrs is not None:
                raise Exception(
                    f"User with user name f{user.username} has multiple entries in LDAP",
                )
            _, attrs = result

        if attrs is not None:
            for key, values in attrs.items():
                try:
                    field = ldap_to_model[key.lower()]
                except KeyError:
                    continue
                (value,) = values
                setattr(user, field, str(value, 'UTF-8', 'ignore'))

    def authenticate(self, request, remote_user):
        """Configure a user after login and return the (potentially) updated user."""

        user = super().authenticate(request, remote_user)
        if user is None:
            return None

        now = datetime.now(tz=timezone.utc)
        expiration = now - timedelta(hours=10)

        def is_up_to_date():
            last_ldap_sync = user.last_ldap_sync
            return last_ldap_sync is not None and last_ldap_sync < expiration

        if is_up_to_date():
            return user

        User = type(user)  # pylint: disable=C0103

        # We're going to try to prevent multiple instances from updating the info
        # simultaneously. This may fail during concurrent access, so retry it,
        # but limit to 3 attempts:
        for _ in range(3):
            try:
                with atomic():
                    user = User.objects.select_for_update().filter(pk=user.pk).get()
                    if is_up_to_date():
                        break

                    user.last_ldap_sync = now
                    try:
                        self._update_user_using_ldap(user)
                    finally:
                        user.save()
                break
            except TransactionManagementError:
                pass
        else:
            # It failed. Give up.
            return None

        return user
