"""Compute results for sets of portfolios or portfolio revisions"""

from math import log, sqrt
from itertools import chain, groupby

from ..stocks.models import AssetPrice


def compute_portfolio_results(portfolios, start_date, end_date):
    """
    Compute results for a set of portfolios.

    Steps:

    1) extract information about each portfolio's selected assets and weights;

    2) query the database for price information for the assets for the
       duration of the game;

    3) for each date for which we have asset price information, for each
       portfolio:

       a) check if we need to switch to a newer portfolio revision

       b) compute the weighted portfolio result for that day

       c) update the total_returns and total_returns_squared values for
          the portfolio

       d) add a result entry for this portfolio for this day.
    """

    class PortfolioState:
        """Passive container to keep state about each portfolio"""

        nobs = 0
        total_returns = 0
        total_returns_squared = 0
        current_revision_state = None

        def __init__(self, **values):
            self.__dict__ = values

    class RevisionState:
        """Passive container to keep state about each portfolio revision"""

        def __init__(self, **values):
            self.__dict__ = values

    # Initialize the portfolio states and portfolio revision states.
    # Each portfolio has a list of revisions, each portfolio revision
    # state has a list of assets with their weights.
    portfolio_states = [
        PortfolioState(
            key=portfolio.pk,
            revision_states=[
                RevisionState(
                    start_date=revision.created_at.astimezone().date(),
                    asset_weights=[
                        (asset.asset_id, asset.weight / 100.0)
                        for asset in revision.portfolioasset_set.all()
                    ],
                )
                for revision in sorted(
                    portfolio.portfoliorevision_set.all(),
                    key=lambda revision: revision.date
                )
            ]
        )
        for portfolio in portfolios
    ]

    # This is where the output will go:
    portfolio_results = {portfolio.pk: [] for portfolio in portfolios}

    # Compute the list of all assets referenced by all revisions, using
    # a set() to deduplicate.
    unique_asset_pks = {
        portfolio_asset.asset_id
        for portfolio_asset in chain.from_iterable(
            revision.portfolioasset_set.all()
            for revision in chain.from_iterable(
                portfolio.portfoliorevision_set.all()
                for portfolio in portfolios
            )
        )
    }

    # This list of asset prices can be a bit large, so we should be careful to
    # not turn this iterator into a list or dict.
    asset_prices = AssetPrice.objects.filter(
        asset__in=unique_asset_pks,
        closing_date__gte=start_date,
        closing_date__lte=end_date,
        closing_price__isnull=False,
    ).order_by('closing_date')

    grouped_asset_prices = groupby(
        asset_prices,
        key=lambda asset_price: asset_price.closing_date,
    )

    _, initial_asset_prices = next(grouped_asset_prices, (None, []))
    previous_logprices = {
        asset_price.asset_id: log(asset_price.closing_price)
        for asset_price in initial_asset_prices
    }

    for date, asset_prices in grouped_asset_prices:
        # Convert the asset prices for this date into a dict for easier
        # lookups:
        todays_asset_logprices = {
            asset_price.asset_id: log(asset_price.closing_price)
            for asset_price in asset_prices
        }

        for portfolio_state in portfolio_states:
            revision_state = portfolio_state.current_revision_state

            revision_states = portfolio_state.revision_states
            while revision_states and revision_states[0].start_date < date:
                # We need to switch to a newer revision for this portfolio:
                revision_state = revision_states.pop(0)
                portfolio_state.current_revision_state = revision_state

            # Check if we already have a valid revision (someone might decide to
            # create a new portfolio after the game has already started).
            if revision_state:
                todays_returns = 0
                at_least_one_asset = False
                for asset_pk, weight in revision_state.asset_weights:
                    try:
                        todays_logprice = todays_asset_logprices[asset_pk]
                        previous_logprice = previous_logprices[asset_pk]
                    except KeyError:
                        pass
                    else:
                        # We have both values so we can include these in the score:
                        at_least_one_asset = True
                        todays_returns += (todays_logprice - previous_logprice) * weight
                if at_least_one_asset:
                    portfolio_state.nobs += 1
                    portfolio_state.total_returns += todays_returns
                    portfolio_state.total_returns_squared += todays_returns * todays_returns
                    current_result = {
                        'date': str(date),
                        'performance': todays_returns,
                    }
                    if portfolio_state.nobs > 1:
                        average_returns = portfolio_state.total_returns / portfolio_state.nobs
                        standard_deviation = sqrt(
                            portfolio_state.total_returns_squared / portfolio_state.nobs -
                            average_returns * average_returns
                        )
                        if standard_deviation:
                            current_result['sharpe_ratio'] = (
                                average_returns / standard_deviation
                            )
                        elif not average_returns:
                            current_result['sharpe_ratio'] = 0.0
                        current_result['risk'] = standard_deviation

                    portfolio_results[portfolio_state.key].append(current_result)

        previous_logprices.update(todays_asset_logprices)

    return portfolio_results


def compute_revision_results(revisions, start_date, end_date):
    """
    Compute results for a set of portfolio revisions.

    Steps:

    1) extract information about each revision's selected assets and weights;

    2) query the database for price information for the assets for the
       duration of the game;

    3) for each date for which we have asset price information, for each
       portfolio revision:

       a) check if this revision has started.
          if not, just copy the day's result from the previous revision
          if so, compute a new result:

       b) compute the weighted portfolio revision result for that day

       c) update the total_returns and total_returns_squared values for
          the revision

       d) add a result entry for this revision for this day.
    """

    class RevisionState:
        """Passive container to keep state about each portfolio revision"""

        nobs = 0
        total_returns = 0
        total_returns_squared = 0

        def __init__(self, **values):
            self.__dict__ = values


    # Initialize the portfolio revision states. Each revision
    # state has a list of assets with their weights.
    revision_states = [
        RevisionState(
            key=revision.pk,
            start_date=revision.created_at.astimezone().date(),
            asset_weights=[
                (asset.asset_id, asset.weight / 100.0)
                for asset in revision.portfolioasset_set.all()
            ],
        )
        for revision in revisions
    ]

    # This is where the output will go:
    revision_results = {revision.pk: [] for revision in revisions}

    # Compute the list of all assets referenced by all revisions, using
    # a set() to deduplicate.
    unique_asset_pks = {
        portfolio_asset.asset_id
        for portfolio_asset in chain.from_iterable(
            revision.portfolioasset_set.all()
            for revision in revisions
        )
    }

    # This list of asset prices can be a bit large, so we should be careful to
    # not turn this iterator into a list or dict.
    asset_prices = AssetPrice.objects.filter(
        asset__in=unique_asset_pks,
        closing_date__gte=start_date,
        closing_date__lte=end_date,
        closing_price__isnull=False,
    ).order_by('closing_date')

    grouped_asset_prices = groupby(
        asset_prices,
        key=lambda asset_price: asset_price.closing_date,
    )

    _, initial_asset_prices = next(grouped_asset_prices, (None, []))
    previous_logprices = {
        asset_price.asset_id: log(asset_price.closing_price)
        for asset_price in initial_asset_prices
    }

    for date, asset_prices in grouped_asset_prices:
        # Convert the asset prices for this date into a dict for easier
        # lookups:
        todays_asset_logprices = {
            asset_price.asset_id: log(asset_price.closing_price)
            for asset_price in asset_prices
        }

        # Keep the most recently computed result so we can reuse it for
        # revisions that haven't started yet.
        current_result = None

        # Inactive revisions must copy their state from the previous one:
        previous_state = None

        for revision_state in revision_states:
            if revision_state.start_date < date:
                todays_returns = 0
                at_least_one_asset = False
                for asset_pk, weight in revision_state.asset_weights:
                    try:
                        todays_logprice = todays_asset_logprices[asset_pk]
                        previous_logprice = previous_logprices[asset_pk]
                    except KeyError:
                        pass
                    else:
                        # We have both values so we can include these in the score:
                        at_least_one_asset = True
                        todays_returns += (todays_logprice - previous_logprice) * weight
                if at_least_one_asset:
                    revision_state.nobs += 1
                    revision_state.total_returns += todays_returns
                    revision_state.total_returns_squared += todays_returns * todays_returns
                    current_result = {
                        'date': str(date),
                        'performance': todays_returns,
                    }
                    if revision_state.nobs > 1:
                        average_returns = revision_state.total_returns / revision_state.nobs
                        standard_deviation = sqrt(
                            revision_state.total_returns_squared / revision_state.nobs -
                            average_returns * average_returns
                        )
                        if standard_deviation:
                            current_result['sharpe_ratio'] = (
                                average_returns / standard_deviation
                            )
                        elif not average_returns:
                            current_result['sharpe_ratio'] = 0.0
                        current_result['risk'] = standard_deviation
            elif previous_state:
                revision_state.nobs = previous_state.nobs
                revision_state.total_returns = previous_state.total_returns
                revision_state.total_returns_squared = previous_state.total_returns_squared

            if current_result is not None:
                revision_results[revision_state.key].append(current_result)

            previous_state = revision_state

        previous_logprices.update(todays_asset_logprices)

    return revision_results

