"""Models for the game app"""

import uuid

from django.db import models
from django.core.validators import MaxValueValidator, MinValueValidator
from django.utils.timezone import now
from django.core.exceptions import ValidationError

from ..stocks.models import Asset, Index
from ..uvt.models import User


class Game(models.Model):
    """A game session"""

    name = models.TextField()
    token = models.UUIDField(default=uuid.uuid4)
    start_date = models.DateField()
    embargo_date = models.DateField(null=True)
    end_date = models.DateField()
    index = models.ForeignKey(to=Index, on_delete=models.PROTECT, null=True)
    assets = models.ManyToManyField(Asset)
    max_portfolios = models.IntegerField(default=3, null=True)
    max_portfolio_revisions = models.IntegerField(default=10, null=True)
    max_portfolio_assets = models.IntegerField(default=4, null=True)


class Participation(models.Model):
    """An instance for each user participating in a game"""

    game = models.ForeignKey(to=Game, on_delete=models.CASCADE)
    user = models.ForeignKey(to=User, on_delete=models.CASCADE)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                name='unique_game',
                fields=['game', 'user'],
            ),
        ]


class Portfolio(models.Model):
    """A portfolio of a user"""

    name = models.TextField()
    participation = models.ForeignKey(to=Participation, on_delete=models.CASCADE)


class PortfolioRevision(models.Model):
    """A revision of a portfolio (created when a user creates or edits a portfolio)"""

    created_at = models.DateTimeField(default=now)
    portfolio = models.ForeignKey(to=Portfolio, on_delete=models.CASCADE)
    date = models.DateField()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                name='unique_date',
                fields=['portfolio', 'date'],
            ),
        ]

def validate_not_zero(value):
    if not value:
        raise ValidationError("Weight must not be zero")

class PortfolioAsset(models.Model):
    """Links an asset to a portfolio revision, with a weight"""

    weight = models.IntegerField(validators=[
        MinValueValidator(-100),
        MaxValueValidator(100),
        validate_not_zero,
    ])
    portfolio_revision = models.ForeignKey(to=PortfolioRevision, on_delete=models.CASCADE)
    asset = models.ForeignKey(to=Asset, on_delete=models.PROTECT)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                name='unique_portfolio_asset',
                fields=['portfolio_revision', 'asset'],
            ),
        ]
