"""Serializers from game models"""

import datetime
from functools import cached_property

from django.core.exceptions import PermissionDenied
from rest_framework.serializers import (
    ModelSerializer,
    ListSerializer,
    IntegerField,
    PrimaryKeyRelatedField,
    ValidationError,
)
from rest_framework.validators import UniqueTogetherValidator

from ..mixins import CreateOnlyFieldSerializerMixin
from ..stocks.models import Asset
from .models import Game, Participation, Portfolio, PortfolioRevision, PortfolioAsset


def validate_end_date(value):
    """
    Validate whether the end date is after the start date.
    """
    start_date = value['start_date']
    end_date = value['end_date']
    if end_date <= start_date:
        raise ValidationError(
            f"The end date ({end_date}) must be after "
            f"the start date ({start_date})",
        )


def validate_embargo_date(value):
    """
    Validate whether the end date is after the start date.
    """

    embargo_date = value.get('embargo_date')
    if embargo_date is None:
        return

    start_date = value['start_date']
    end_date = value['end_date']
    if embargo_date <= start_date:
        raise ValidationError(
            f"The embargo date ({embargo_date}) must be after "
            f"the start date ({start_date})",
        )

    if embargo_date >= end_date:
        raise ValidationError(
            f"The embargo date ({end_date}) must be before "
            f"the end date ({end_date})",
        )


class GameSerializer(ModelSerializer):
    """Serializer for game model"""

    assets = PrimaryKeyRelatedField(
        queryset=Asset.objects.all(),
        many=True,
        allow_empty=True,
    )

    class Meta:
        model = Game
        read_only_fields = ['pk']
        fields = read_only_fields + [
            'name',
            'token',
            'start_date',
            'embargo_date',
            'end_date',
            'index',
            'assets',
            'max_portfolios',
            'max_portfolio_revisions',
            'max_portfolio_assets',
        ]
        validators = [
            validate_end_date,
            validate_embargo_date,
        ]


class ParticipationSerializer(CreateOnlyFieldSerializerMixin, ModelSerializer):
    """Serializer for participation model"""

    class Meta:
        model = Participation
        read_only_fields = ['pk']
        create_only_fields = ['game', 'user']
        fields = read_only_fields + create_only_fields
        validators = [
            UniqueTogetherValidator(
                queryset=Participation.objects.all(),
                fields=['game', 'user'],
                message="This user already participates in this game.",
            )
        ]


class PortfolioAssetSerializer(ModelSerializer):
    """
    Serializer for portfolio asset model.

    This serializer is tailored towards use by PortfolioRevisionSerializer
    and should probably not be used directly by a ViewSet.
    """

    class Meta:
        model = PortfolioAsset
        fields = ['asset', 'weight']


class LastRevisionTodayValidator:
    """
    If the game has started and the last revision was created today, no
    further revisions are possible.
    """

    requires_context = True

    def __call__(self, value, serializer):
        today = datetime.date.today()
        if serializer.context['view'].action == 'create':
            portfolio = value['portfolio']
            if today >= portfolio.participation.game.start_date:
                if any(
                    revision.date >= today
                    for revision in portfolio.portfoliorevision_set.all()
                ):
                    raise ValidationError("You can't update twice on the same day")
        else:
            if today >= serializer.instance.portfolio.participation.game.start_date:
                raise ValidationError("You can't update revisions after the game has started")


class LastRevisionDifferentValidatorImplementation:
    def __init__(self, value, serializer):
        self.value = value
        self.serializer = serializer

    def __call__(self):
        if self.serializer.context['view'].action == 'create':
            self.check_create_action()

    def check_create_action(self):
        if datetime.date.today() >= self.game.start_date:
            self.check_last_revision_different()

    @cached_property
    def game(self):
        return self.portfolio.participation.game

    @cached_property
    def portfolio(self):
        return self.value['portfolio']

    def check_last_revision_different(self):
        """
        Get the last revision (if any) by looping over a queryset
        that is either empty or contains one element:
        """

        for last_revision in self.last_revision_set:
            self.check_revision_different(last_revision)

    @cached_property
    def last_revision_set(self):
        """A queryset containing either the last revision or nothing"""

        return (
            self.portfolio
            .portfoliorevision_set
            .order_by('-date')
            [:1]
        )

    def check_revision_different(self, revision):
        """
        Check if the new set of assets and weights is
        identical to this revision, and balk if it is:
        """

        new_assets = self.assets_data_as_dict
        last_revision_assets = self.get_revision_assets_as_dict(revision)

        if new_assets == last_revision_assets:
            # Nothing changed, which is invalid:
            raise ValidationError(
                "The new revision must differ from the previous one"
            )

    @cached_property
    def assets_data_as_dict(self):
        return {
            asset['asset'].pk: asset['weight']
            for asset in self.value['portfolioasset_set']
        }

    def get_revision_assets_as_dict(self, revision):
        return {
            asset.asset_id: asset.weight
            for asset in revision.portfolioasset_set.all()
        }


class LastRevisionDifferentValidator:
    """
    A proposed new revision should at least be different from the
    previous one, otherwise it's just a waste of revisions.
    """

    requires_context = True

    def __call__(self, value, serializer):
        implementation = LastRevisionDifferentValidatorImplementation(value, serializer)
        implementation()


class MaxPortfolioRevisionNumberValidator:
    """
    Checks the current number of portfolio revisions against the maximum number
    configured for the game it belongs to.

    If the game hasn't started yet the number of existent revisions doesn't
    matter as they will all be replaced anyway.
    """

    requires_context = True

    def __call__(self, value, serializer):
        if serializer.context['view'].action == 'create':
            portfolio = value['portfolio']
            game = portfolio.participation.game
            if datetime.date.today() < game.start_date:
                return

            portfolio_revision_count = portfolio.portfoliorevision_set.count()
            max_portfolio_revisions = game.max_portfolio_revisions
            if (
                max_portfolio_revisions is not None
                and portfolio_revision_count > max_portfolio_revisions
            ):
                raise ValidationError(
                    f"This game only allows for {max_portfolio_revisions} portfolio "
                    f"revisions and {portfolio_revision_count} are already present",
                )


class MaxPortfolioAssetNumberValidator:
    """
    Checks the current number of portfolio assets of a portfolio revision against
    the maximum number configured for the game it belongs to.
    """

    requires_context = True

    def __call__(self, value, serializer):
        if serializer.context['view'].action == 'create':
            portfolio = value['portfolio']
        else:
            portfolio = serializer.instance.portfolio

        portfolio_asset_count = len(value['portfolioasset_set'])
        max_portfolio_assets = portfolio.participation.game.max_portfolio_assets
        if max_portfolio_assets is not None and portfolio_asset_count > max_portfolio_assets:
            raise ValidationError(
                f"This game only allows for {max_portfolio_assets} assets in "
                f"a portfolio and you submitted {portfolio_asset_count}",
            )


class PortfolioRevisionGameAssetValidator:
    """
    Checks whether the portfolio assets of a portfolio revision are valid
    for this particular game.
    """

    requires_context = True

    def __call__(self, value, serializer):
        portfolio = value.get('portfolio')
        if portfolio is None:
            portfolio = serializer.instance.portfolio
            if portfolio is None:
                return
        portfolio_asset_pks = set(
            portfolio_asset['asset'].pk
            for portfolio_asset in value['portfolioasset_set']
        )
        game_asset_pks = set(
            portfolio.participation.game.assets.values_list(
                'pk',
                flat=True,
            ).all()
        )
        invalid_asset_pks = portfolio_asset_pks - game_asset_pks
        if invalid_asset_pks:
            raise ValidationError(
                f"Portfolio assets {invalid_asset_pks!r} "
                f"are not valid for this game ({game_asset_pks!r}).",
            )


class PortfolioRevisionGameEndValidator:
    """
    Modifications do not make sense after the game has ended.
    """

    requires_context = True

    def __call__(self, value, serializer):
        if serializer.context['view'].action == 'create':
            portfolio = value['portfolio']
        else:
            portfolio = serializer.instance.portfolio
        game = portfolio.participation.game
        if datetime.date.today() >= (game.embargo_date or game.end_date):
            raise ValidationError("This game has already ended")


class PortfolioRevisionSerializer(CreateOnlyFieldSerializerMixin, ModelSerializer):
    """Serializer for portfolio revision model"""

    assets = PortfolioAssetSerializer(source='portfolioasset_set', many=True)

    class Meta:
        model = PortfolioRevision
        read_only_fields = ['pk', 'created_at', 'date']
        create_only_fields = ['portfolio', 'assets']
        fields = read_only_fields + create_only_fields
        validators = [
            LastRevisionTodayValidator(),
            LastRevisionDifferentValidator(),
            MaxPortfolioRevisionNumberValidator(),
            MaxPortfolioAssetNumberValidator(),
            PortfolioRevisionGameAssetValidator(),
            PortfolioRevisionGameEndValidator(),
        ]

    def validate_portfolio(self, portfolio):
        user = self.context['request'].user
        if user.is_admin or portfolio.participation.user == user:
            return portfolio
        raise PermissionDenied("You must be an owner of this portfolio")

    def validate_assets(self, assets_data):
        """Validate aspects of portfolio assets that can only be validated in combination"""

        if len(set(asset_data['asset'].pk for asset_data in assets_data)) < len(assets_data):
            raise ValidationError("An asset may only be referenced once in a portfolio revision")

        if sum(asset_data['weight'] for asset_data in assets_data) != 100:
            raise ValidationError("The sum of weights of a portfolio revision must be 100")

        return assets_data

    def create(self, validated_data):
        """
        Create a new portfolio revision by adding a PortfolioAsset for
        each asset in the input data.
        """

        today = datetime.date.today()
        portfolio = validated_data['portfolio']
        portfolio_assets_data = validated_data['portfolioasset_set']
        revisions = portfolio.portfoliorevision_set

        if today < portfolio.participation.game.start_date:
            # The game hasn't even started yet, so just delete all existent
            # revisions and let the code below create a new one.
            revisions.all().delete()

        revision = PortfolioRevision.objects.create(
            portfolio=portfolio,
            date=today,
        )

        for portfolio_asset_data in portfolio_assets_data:
            portfolio_asset = PortfolioAsset.objects.create(
                portfolio_revision=revision,
                **portfolio_asset_data,
            )
            portfolio_asset.save()

        return revision

    def update(self, instance, validated_data):
        """
        Update this portfolio revision by deleting all old asset references
        and adding in newly created ones.
        """

        for portfolio_asset in instance.portfolioasset_set.all():
            portfolio_asset.delete()

        for portfolio_asset_data in validated_data['portfolioasset_set']:
            portfolio_asset = PortfolioAsset.objects.create(
                portfolio_revision=instance,
                **portfolio_asset_data,
            )
            portfolio_asset.save()

        return instance


class PortfolioGameStartValidator:
    """
    Create/remove does not make sense after the game has started.
    """

    requires_context = True

    def __call__(self, value, serializer):
        action = serializer.context['view'].action
        if action in {'create', 'delete'}:
            if datetime.date.today() < value['participation'].game.start_date:
                return value

            if action == 'delete':
                # Empty portfolios are pointless, so make an exception:
                if serializer.instance.portfoliorevision_set.count() == 0:
                    return value

            raise ValidationError("This game has already started")

        return value


class PortfolioGameEndValidator:
    """
    Modifications do not make sense after the game has ended.
    """

    requires_context = True

    def __call__(self, value, serializer):
        if serializer.context['view'].action == 'create':
            participation = value['participation']
        else:
            participation = serializer.instance.participation
        game = participation.game
        if datetime.date.today() < (game.embargo_date or game.end_date):
            return value

        raise ValidationError("This game has already ended")


class MaxPortfolioNumberValidator:
    """
    Checks the current number of portfolios against the maximum number
    configured for the game it belongs to.
    """

    requires_context = True

    def __call__(self, value, serializer):
        if serializer.context['view'].action == 'create':
            participation = value['participation']
            portfolio_count = participation.portfolio_set.count()
            max_portfolios = participation.game.max_portfolios
            if max_portfolios is not None and portfolio_count >= max_portfolios:
                raise ValidationError(
                    f"This game only allows for {max_portfolios} portfolios "
                    f"and {portfolio_count} are already present",
                )

        return value


class SingletonSerializer(ListSerializer):
    def to_representation(self, data):
        as_list = super().to_representation(data)
        if as_list:
            singleton, = as_list
            return singleton
        else:
            return None

    def create(self, validated_data):
        raise NotImplementedError("Singleton.create() is not implemented")


class PortfolioSerializer(CreateOnlyFieldSerializerMixin, ModelSerializer):
    """Serializer for portfolio model"""

    def validate_participation(self, participation):
        user = self.context['request'].user
        if user.is_admin or participation.user == user:
            return participation
        raise PermissionDenied("You must be an owner of this participation")

    last_revision = SingletonSerializer(
        child=PortfolioRevisionSerializer(),
        read_only=True,
    )

    revision_count = IntegerField(read_only=True)

    class Meta:
        model = Portfolio
        read_only_fields = ['pk', 'last_revision', 'revision_count']
        create_only_fields = ['participation']
        fields = read_only_fields + create_only_fields + ['name']
        validators = [
            MaxPortfolioNumberValidator(),
            PortfolioGameStartValidator(),
            PortfolioGameEndValidator(),
        ]
