"""Views for Game related endpoints"""

from uuid import UUID
from itertools import chain, groupby
from datetime import date

from openpyxl import Workbook as ExcelWorkbook
from openpyxl.writer.excel import save_virtual_workbook

from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.models import OuterRef, Prefetch, Subquery, Count
from django.http import HttpResponse, HttpResponseNotFound

from rest_framework.views import APIView
from rest_framework.permissions import BasePermission, SAFE_METHODS
from rest_framework.viewsets import ModelViewSet
from rest_framework.response import Response
from rest_framework.serializers import ValidationError
from rest_framework_extensions.mixins import NestedViewSetMixin
from rest_framework.decorators import action

from ..mixins import UpdateableNestedViewSetMixin, UserFilterNestedViewSetMixin
from .models import Game, Participation, Portfolio, PortfolioRevision
from .serializers import (
    GameSerializer,
    ParticipationSerializer,
    PortfolioSerializer,
    PortfolioRevisionSerializer,
)
from .compute import (
    compute_portfolio_results,
    compute_revision_results,
)


class GamePermission(BasePermission):
    message = "You are not an admin."

    def has_permission(self, request, view):
        return request.user.is_admin or request.method in SAFE_METHODS


class GameDownloadPermission(BasePermission):
    message = "You are not an admin."

    def has_permission(self, request, view):
        return request.user.is_admin


class GameViewSet(
    UserFilterNestedViewSetMixin,
    NestedViewSetMixin,
    ModelViewSet,
):
    """API endpoint"""

    queryset = Game.objects.all()

    serializer_class = GameSerializer
    filterset_fields = ['name']
    search_fields = ['name']
    permission_classes = [GamePermission]
    parent_user_field = 'participation__user'

    @action(detail=True, permission_classes=[GameDownloadPermission])
    def download(self, request, pk, _=None):
        """Admin downloads all portfolio information"""

        try:
            game = Game.objects.get(pk=pk)
        except Game.DoesNotExist:
            return HttpResponseNotFound(b'Game not found')

        assets = list(game.assets.order_by('ticker'))

        last_revision = (
            PortfolioRevision.objects
            .filter(portfolio_id=OuterRef('portfolio_id'))
            .order_by('-date')
        )

        portfolios = (
            Portfolio.objects.prefetch_related(
                Prefetch(
                    'portfoliorevision_set',
                    queryset=PortfolioRevision.objects.filter(
                        pk=Subquery(last_revision.values('pk')[:1]),
                    ),
                    to_attr='last_revision',
                ),
                'last_revision__portfolioasset_set',
                'participation',
                'participation__user',
            )
            .filter(participation__game=game)
            .order_by('participation__user__full_name', 'pk')
        )

        workbook = ExcelWorkbook(write_only=True)
        workbook.create_sheet()
        worksheet = workbook.active
        worksheet.append([
            'Student',
            'SNR',
            'Portfolio',
            *(asset.ticker for asset in assets),
        ])
        for portfolio in portfolios:
            user = portfolio.participation.user
            for revision in portfolio.last_revision:
                revision_weights = {
                    asset.asset.pk: asset.weight
                    for asset in revision.portfolioasset_set.all()
                }
                row_assets = [
                    revision_weights.get(asset.pk, 0) for asset in assets
                ]
                worksheet.append([
                    user.full_name,
                    user.student_number,
                    portfolio.name,
                    *row_assets,
                ])

        return HttpResponse(
            save_virtual_workbook(workbook),
            content_type='application/vnd'
            '.openxmlformats-officedocument.spreadsheetml.sheet',
        )

    @action(detail=True)
    def results(self, request, pk, _=None):
        """Compute results of all portfolios of all participants of a game"""
        try:
            game = self.get_queryset().prefetch_related(
                'participation_set',
                'participation_set__portfolio_set',
                'participation_set__portfolio_set__portfoliorevision_set',
                'participation_set__portfolio_set__portfoliorevision_set__portfolioasset_set',
                'participation_set__portfolio_set__portfoliorevision_set__portfolioasset_set__asset',
                'participation_set__user',
            ).get(pk=pk)
        except Game.DoesNotExist:
            return HttpResponseNotFound(b'Game not found')

        portfolios = list(chain.from_iterable(
            participation.portfolio_set.all()
            for participation in game.participation_set.all()
        ))

        # Only consider the embargo date if the user is not an admin:
        end_date = not request.user.is_admin and game.embargo_date or game.end_date

        try:
            date_parameter = request.GET['date']
        except KeyError:
            pass
        else:
            try:
                user_end_date = date.fromisoformat(date_parameter)
            except:
                raise ValidationError("invalid date parameter") from None
            if user_end_date < game.start_date or user_end_date > end_date:
                raise ValidationError("invalid date parameter")
            end_date = user_end_date

        leaderboard = []

        results = compute_portfolio_results(portfolios, game.start_date, end_date)
        for (user, portfolios) in groupby(
            portfolios,
            lambda portfolio: portfolio.participation.user
        ):
            portfolioboard = []
            for portfolio in portfolios:
                pk = portfolio.pk
                portfolioscore = {
                    'pk': pk,
                    'name': portfolio.name,
                }
                portfolio_results = results[pk]
                if portfolio_results:
                    portfolioscore['current'] = portfolio_results[-1]
                    if len(portfolio_results) > 1:
                        portfolioscore['previous'] = portfolio_results[-2]
                portfolioboard.append(portfolioscore)

            userboard = {
                'pk': user.pk,
                'portfolios': portfolioboard,
            }
            user_name = user.full_name
            if user_name is not None:
                userboard['name'] = user_name
            leaderboard.append(userboard)

        return Response(leaderboard)


class JoinView(APIView):
    """User joins the game."""

    def post(self, request, _=None):
        """Check game token and enroll user."""

        token = request.data.get('token')
        if not isinstance(token, str):
            raise ValidationError("missing or invalid token")

        try:
            uuid = UUID(token)
        except ValueError:
            raise ValidationError("missing or invalid token")

        if str(uuid) != token:
            raise ValidationError("missing or invalid token")

        try:
            game = Game.objects.get(token=uuid)
        except (DjangoValidationError, Game.DoesNotExist):
            raise ValidationError("missing or invalid token")

        # Create the participation object if it doesn't exist
        participation, _ = Participation.objects.get_or_create(
            game=game,
            user=request.user,
        )

        return Response(ParticipationSerializer(participation).data)


class ParticipationPermission(BasePermission):
    message = "You are not an admin."

    def has_permission(self, request, view):
        return request.user.is_admin or request.method in SAFE_METHODS


class ParticipationViewSet(
    UpdateableNestedViewSetMixin,
    UserFilterNestedViewSetMixin,
    NestedViewSetMixin,
    ModelViewSet,
):
    """API endpoint"""

    queryset = Participation.objects.all()

    serializer_class = ParticipationSerializer
    permission_classes = [ParticipationPermission]
    filterset_fields = ['user', 'game']
    parent_user_field = 'user'

    @action(detail=True)
    def results(self, request, pk, _=None):
        """Compute results of all portfolios of a participation"""
        try:
            participation = self.get_queryset().select_related(
                'game',
            ).get(pk=pk)
        except Participation.DoesNotExist:
            return HttpResponseNotFound(b'Participation not found')

        game = participation.game

        portfolios = list(
            participation
            .portfolio_set
            .prefetch_related(
                'portfoliorevision_set',
                'portfoliorevision_set__portfolioasset_set',
                'portfoliorevision_set__portfolioasset_set__asset',
            )
        )

        # Only consider the embargo date if the user is not an admin:
        end_date = not request.user.is_admin and game.embargo_date or game.end_date

        return Response(compute_portfolio_results(portfolios, game.start_date, end_date))


class PortfolioPermission(BasePermission):
    def has_object_permission(self, request, view, obj):
        user = request.user
        return user.is_admin or obj.participation.user == user


class PortfolioViewSet(
    UpdateableNestedViewSetMixin,
    UserFilterNestedViewSetMixin,
    NestedViewSetMixin,
    ModelViewSet,
):
    """API endpoint"""

    queryset = Portfolio.objects.all()

    serializer_class = PortfolioSerializer
    permission_classes = [PortfolioPermission]
    filterset_fields = ['name', 'participation']
    search_fields = ['name']
    parent_user_field = 'participation__user'

    def get_queryset(self):
        queryset = super().get_queryset()

        last_revision = (
            PortfolioRevision.objects
            .filter(portfolio_id=OuterRef('portfolio_id'))
            .order_by('-date')
        )

        return queryset.all().select_related(
            'participation',
            'participation__game',
        ).prefetch_related(
            Prefetch(
                'portfoliorevision_set',
                queryset=PortfolioRevision.objects.filter(
                    pk=Subquery(last_revision.values('pk')[:1]),
                ),
                to_attr='last_revision',
            ),
            'last_revision__portfolioasset_set',
        ).annotate(
            revision_count=Count('portfoliorevision'),
        )

    @action(detail=True)
    def results(self, request, pk, _=None):
        """Compute results of all revisions of a portfolio"""
        try:
            portfolio = self.get_queryset().select_related(
                'participation',
                'participation__game',
            ).get(pk=pk)
        except Portfolio.DoesNotExist:
            return HttpResponseNotFound(b'Portfolio not found')

        game = portfolio.participation.game

        revisions = list(
            portfolio
            .portfoliorevision_set
            .order_by('portfolio', 'date')
            .prefetch_related(
                'portfolioasset_set',
                'portfolioasset_set__asset',
            )
        )

        # Only consider the embargo date if the user is not an admin:
        end_date = not request.user.is_admin and game.embargo_date or game.end_date

        return Response(compute_revision_results(revisions, game.start_date, end_date))


class PortfolioRevisionPermission(BasePermission):
    def has_object_permission(self, request, view, obj):
        if request.method == 'DELETE':
            return False
        user = request.user
        return user.is_admin or obj.portfolio.participation.user == user


class PortfolioRevisionViewSet(
    UpdateableNestedViewSetMixin,
    UserFilterNestedViewSetMixin,
    NestedViewSetMixin,
    ModelViewSet,
):
    """API endpoint"""

    queryset = PortfolioRevision.objects.all()

    serializer_class = PortfolioRevisionSerializer
    permission_classes = [PortfolioRevisionPermission]
    filterset_fields = ['portfolio']
    parent_user_field = 'portfolio__participation__user'

    def get_queryset(self):
        return super().get_queryset().select_related(
            'portfolio',
            'portfolio__participation',
            'portfolio__participation__game',
        ).prefetch_related(
            'portfolio__portfoliorevision_set',
        )

    @action(detail=True)
    def results(self, request, pk, _=None):
        """Compute results of a specific revision of a portfolio"""
        try:
            revision = self.get_queryset().select_related(
                'portfolio',
                'portfolio__participation',
                'portfolio__participation__game',
            ).get(pk=pk)
        except PortfolioRevision.DoesNotExist:
            return HttpResponseNotFound(b'Portfolio revision not found')

        portfolio = revision.portfolio
        game = portfolio.participation.game

        revisions = list(
            portfolio
            .portfoliorevision_set
            .filter(date__lte=revision.date)
            .order_by('portfolio', 'date')
            .prefetch_related(
                'portfolioasset_set',
                'portfolioasset_set__asset',
            )
        )

        # Only consider the embargo date if the user is not an admin:
        end_date = not request.user.is_admin and game.embargo_date or game.end_date

        return Response(
            compute_revision_results(revisions, game.start_date, end_date)
            .get(revision.pk, [])
        )
