"""Some general-purpose mixins"""

from django.core.exceptions import (
    PermissionDenied as DjangoPermissionDeniedError,
)

_EMPTY_LIST = []


class CreateOnlyFieldSerializerMixin:
    """Mixin to enable the create_only_fields Meta attribute"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        context = self.context
        # context['view'] may not exist when the HTML DRF frontend is querying
        # the attributes to generate HTML form fields for.
        if 'view' in context and context['view'].action != 'create':
            OldMeta = self.Meta

            old_read_only_fields = getattr(OldMeta, 'read_only_fields', _EMPTY_LIST)
            old_create_only_fields = getattr(OldMeta, 'create_only_fields', _EMPTY_LIST)

            class Meta(OldMeta):
                """Meta with adjusted read_only_fields"""
                read_only_fields = list({*old_read_only_fields, *old_create_only_fields})

            self.Meta = Meta


class UserFilterNestedViewSetMixin:
    """
    Mixin for nested ViewSets that filters based on the logged in user.
    Admins get unfiltered responses.
    """

    def get_parents_query_dict(self):
        """
        Hook into get_parents_query_dict() so we can access the filter
        that is being built. We can't simply hook get_queryset() and add
        a .filter() to the queryset because
        .filter(participation__user=foo, participation__user=bar)
        means something else than:
        .filter(participation__user=foo).filter(participation__user=bar)
        and we need the semantics of the former.

        See https://docs.djangoproject.com/en/2.2/topics/db/queries/#spanning-multi-valued-relationships
        for more details.

        Add a parent_user_filter property to specify which parent query filter
        should contain the desired user.

        Insert this mixin before NestedViewSetMixin.
        """

        query_dict = super().get_parents_query_dict()
        user = self.request.user
        if not user.is_admin:
            pk = str(user.pk)
            if query_dict.setdefault(self.parent_user_field, pk) != pk:
                raise DjangoPermissionDeniedError(
                    "You must be an admin to access other people's data"
                )
        return query_dict


class UpdateableNestedViewSetMixin:
    """Mixin for nested ViewSets that adds in parameters from the URL path"""

    def get_serializer(self, *args, **kwargs):
        """Override for get_serializer() that adds parameters from the path"""

        try:
            data = kwargs.pop('data')
        except KeyError:
            pass
        else:
            # The names in this dict come from the parents_query_lookups
            # parameter to nested routers.
            # For a hypothetical request like this:
            #    /games/3/participations/5/portfolios/21
            # parents_query_dict would look like this:
            #    {'game': 3, 'participation': 5, 'portfolio': 21}
            parents_query_dict = self.get_parents_query_dict()

            return super().get_serializer(*args, data={**parents_query_dict, **data}, **kwargs)

        return super().get_serializer(*args, **kwargs)


UpdatableNestedViewSetMixin = UpdateableNestedViewSetMixin
