"""Session related views"""

from rest_framework.views import APIView
from rest_framework.viewsets import ModelViewSet
from rest_framework.response import Response
from rest_framework.permissions import BasePermission, SAFE_METHODS
from rest_framework_extensions.mixins import NestedViewSetMixin
from .serializers import UserSerializer
from .models import User


class UserPermission(BasePermission):
    message = "User information is read-only."

    def has_permission(self, request, view):
        return request.method in SAFE_METHODS


class UserViewSet(NestedViewSetMixin, ModelViewSet):
    """API endpoint"""

    queryset = User.objects.all()

    serializer_class = UserSerializer
    filterset_fields = ['username', 'anr', 'student_number']
    search_fields = ['full_name', '=username', '=anr', '=student_number']
    permission_classes = [UserPermission]

    def get_queryset(self):
        queryset = super().get_queryset()
        user = self.request.user
        if user.is_admin:
            return queryset
        else:
            return queryset.filter(pk=user.pk)


class SessionView(APIView):
    """Information pertaining to the currently logged in user."""

    def get(self, request, _=None):
        """Return information pertaining to the currently logged in user."""

        res = {'user': UserSerializer(request.user).data}
        csrf_token = request.META.get('CSRF_COOKIE')
        if csrf_token is not None:
            res['csrf'] = csrf_token

        return Response(res)
